import { ConnectionManager as ConnectionManagerInterface, AdapterSet, TransportService, TransportEvent, Transport } from "../interface/connector.interface"
import { TransmitterConnectionAdapter } from './connector.transmitter'
import { ReceiverConnectionAdapter } from './connector.receiver'
import { RequestResponseConnectionAdapter } from './connector.request.response'
import { v4 as uuidv4 } from 'uuid'
import { Subject } from "rxjs"
import { WebsocketTransportService } from "../transport/websocket"
import { HttpTransportService } from "../transport/http"
import config from '../config/config.json';

export class ConnectionManager implements ConnectionManagerInterface {
    private transportServiceArray: TransportService[] = []
    private transportSet: Set<TransportSet> = new Set()
    private adapterSet: AdapterSet[] = []
    private event!: Subject<TransportEvent>

    constructor(event: Subject<TransportEvent>, browserEnv?: boolean | undefined) {
        this.event = event
        console.log(`Connection Manager: Contructing ConnectionManager....`)

        this.sort(this.transportSet)
        this.transportSet.forEach(set => {
            this.setUpTransportService(set, event, browserEnv)
        })
    }

    async getAdapter(clientId: string): Promise<AdapterSet> {
        return new Promise((resolve, reject) => {
            console.log(`Instantiating an adapter set....`)
            let transportType: Transport = Transport.Websocket // as default  for now
            let adapterId: string = clientId
            let transportService: TransportService | undefined = this.transportServiceArray.find(obj => obj.getInfo() == transportType)
            if (transportService) {
                let transmitterAdapter: TransmitterConnectionAdapter = new TransmitterConnectionAdapter(adapterId, transportType, transportService)
                let receiverAdapter: ReceiverConnectionAdapter = new ReceiverConnectionAdapter(adapterId, transportType, transportService)
                let adapterSet: AdapterSet = {
                    id: adapterId,
                    dateCreated: new Date(),
                    transmitterAdapter: transmitterAdapter,
                    receiverAdapter: receiverAdapter,
                    requestResponsAdapter: new RequestResponseConnectionAdapter(transmitterAdapter, receiverAdapter)
                }
                this.adapterSet.push(adapterSet)

                this.event.next({
                    id: uuidv4(),
                    event: 'New Adapter',
                    data: adapterId
                } as TransportEvent)
                resolve(adapterSet)
            } else {
                reject(`Connection Manager: No Transport Service Instantiated!`)
            }
        })
    }

    public getTransportArray(): TransportService[] {
        return this.transportServiceArray
    }


    // Server to be set up as well as acquiring client information if needed. Like in the case for grpc and socket. Http not requ`ired.
    private setUpTransportService(transportSet: TransportSet, event: Subject<TransportEvent>, isClient?: boolean): void {
        this.instantiateTransportService(transportSet.transport, event).then((transportService: TransportService) => {
            this.transportServiceArray.push(transportService)
            if (transportService instanceof WebsocketTransportService) {
                console.log(`Just Double Checking... this is websocket`)
                if (isClient) {
                    // please note this is subject to change depending on the UI environemnt. Angular has their own built in function to read json file based on Swopt-UI
                    transportService.startClient(config.connection.transmitter)
                } else {
                    transportService.startServer(transportSet.port);
                }
            } else if (transportService instanceof HttpTransportService) {
                console.log(`Just Double Checking... this is http`)
                // Additional Http-specific setup if needed.
                if (isClient) {
                    transportService.startClient(`http://localhost:3000`)
                } else {
                    transportService.startServer(transportSet.port)
                }
            }
        }).catch((error) => { throw new Error(error) })
    }

    private async instantiateTransportService(transportType: Transport, event: Subject<TransportEvent>): Promise<TransportService> {
        return new Promise((resolve, reject) => {
            if (transportType == Transport.Websocket) {
                resolve(new WebsocketTransportService(event))
            }
            else if (transportType == Transport.Http) {
                resolve(new HttpTransportService(event))
            } else {
                reject(`No Transport Service is not properly instantiated`)
            }
        })
    }

    private sort(transportSet: Set<TransportSet>): void {
        let transportList: string[] = process.env.Transport?.split(',') || []
        let portList: number[] = (process.env.PORT?.split(',') || []).map(port => Number(port));
        transportList.forEach((transport, index) => {
            transportSet.add({ transport: transport, port: portList[index] } as unknown as TransportSet)
        })
        console.log(this.transportSet)
    }

}


interface TransportSet {
    transport: Transport,
    port: number
}

