import { Express } from 'express';
import { Observable, Subject } from "rxjs";
import { ClientObject, Transport, TransportEvent, TransportMessage, TransportService } from "../interface/connector.interface";
import { v4 as uuidv4 } from 'uuid'
import config from '../config/config.json';
import { handleClientHttpConnection, handleHttpClient, initiateClientToServer, startHttpServer } from "../utils/http.utils";
import { WrappedMessage } from '../utils/message.ordering';
import { error } from 'console';

export class HttpTransportService implements TransportService {
    private baseUrl!: string;
    private info: Transport = Transport.Http
    private connectedHttpServer: ConnectedHttpServer[] = [] // to allow the possibility of having to communicate with multiple servers as a client
    private connectedHttpClients: ConnectedHttpClient[] = [] // to keep track of the all the clients that are connected
    transportEvent!: Subject<TransportEvent>

    constructor(event: Subject<TransportEvent>) {
        this.baseUrl = config.connection.transmitter
        this.transportEvent = event
    }

    public getInfo(): Transport {
        return this.info
    }

    public emit(message: TransportMessage): void {
        let clientObj: ConnectedHttpClient | undefined = this.connectedHttpClients.find(obj => obj.id == message.target)
        let serverObj: ConnectedHttpServer | undefined = this.connectedHttpServer.find(obj => obj.id === message.target)

        // for server usage
        if (clientObj && clientObj.connectionState.getValue() == 'ONLINE') {
            clientObj.instance.emit(`message`, message.payload)
        }
        // for client usage
        if (serverObj && serverObj.connectionState.getValue() == 'ONLINE') {
            fetch(`${this.baseUrl}/message`, {
                method: 'POST',
                headers: { 'Content-Type': 'application/json' },
                body: JSON.stringify(message),
            }).catch((error) => console.error('HTTP emit error:', error));
        }

    }

    public subscribe(): Observable<TransportEvent> {
        return this.transportEvent.asObservable();
    }

    public startServer(port: number): void {
        startHttpServer(port).subscribe({
            next: (client: ConnectedHttpClient) => {
                handleHttpClient(client, this.connectedHttpClients).subscribe({
                    next: event => this.transportEvent.next(event),
                    error: error => console.error(error),
                    complete: () => (`Client ${client.id} disconnected...`)
                })
            },
            error: error => console.error(error),
            complete: () => console.log(`...`)
        })
    }

    public startClient(url: string): void {
        initiateClientToServer(url, this.transportEvent, this.connectedHttpServer).then((connectedHttpServer: ConnectedHttpServer) => {
            handleClientHttpConnection(url, connectedHttpServer).subscribe(this.transportEvent)
        }).catch((error) => console.error(`HttpTransport ERROR:`, error))
    }

}


export interface ConnectedHttpClient extends ClientObject {
    instance: Express
    responseStream: Subject<WrappedMessage>
}

export interface ConnectedHttpServer extends ClientObject {
}